# -*- coding: UTF-8 -*-
# -*- Mode: Python; py-indent-offset: 4 -*-
# * Copyright (C) Nikolay Kim <fafhrd@legco.biz> 

import traceback, time, sys
from types import IntType, ListType, NoneType, TupleType, StringType

from zLOG import LOG, INFO
from DateTime import DateTime
from Acquisition import aq_base
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, Permissions
from OFS.SimpleItem import SimpleItem
from ZODB.PersistentMapping import PersistentMapping
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.TimerService import getTimerService
from App.Dialogs import MessageDialog

from utils import UniqueObject
from zLOG import LOG, INFO, WARNING, ERROR

from permissions import *
from Products.ZopeScheduler.ISchedulingTask import ISchedulingTask

INTERVAL = 10
TIME_TYPE_DEFS = {
    'minute':   {'name': 'Minute',          'min': 0, 'max': 60, 'fraction': 10, 'unit': 'min'},
    'hour':     {'name': 'Hour',            'min': 0, 'max': 23, 'unit': 'hour'},
    'dom':      {'name': 'Day of Month',    'min': 1, 'max': 31, 'unit': 'day'},
    'dow':      {'name': 'Day of Week',     'min': 1, 'max': 7,  'unit': 'day'},
    'month':    {'name': 'Month',           'min': 1, 'max': 12, 'unit': 'month'}
}


def manage_addZopeScheduler(self, REQUEST=None):
    """ """
    id = ZopeScheduler.id
    sched = ZopeScheduler()
    self._setObject(id, sched)
    ms = self._getOb(id)

    if REQUEST:
        return self.manage_main(self, REQUEST, update_menu=1)

class SchedulingError (Exception):
    """Error while attempting to stage an object"""

class ZopeScheduler(UniqueObject, SimpleItem):
    """ """
    meta_type = 'Zope Scheduler'

    id = 'zope_scheduler'

    security = ClassSecurityInfo()

    security.declareProtected( Permissions.view_management_screens, 'manage_viewSubscriptions')
    manage_viewSubscriptions = PageTemplateFile( 'zpt/view_subscriptions', globals() )

    manage_options = (
        ({'label': 'Scheduler', 'action':'manage_viewSubscriptions'},) +
        SimpleItem.manage_options)

    def __init__(self):
        self._tasks = PersistentMapping()
        self._methods = PersistentMapping()

    security.declareProtected( ScheduleTasks, 'registerTask' )
    def registerTask(self, path, **time_args):
        """ register a task for execution
            a task is a callable object (python script, method et al)
            valid kwargs keys are
            o minute: 0-60
            o hour: 0-24
            o dom (day of month): 1-31
            o dow (day of week): 1-7
            o month: 1-12
        """
        time_props = self._checkTimeArgs(self._parseTimeArgs(**time_args))
        self._tasks[path] = time_props

    security.declarePrivate( '_parseTimeArgs' )
    def _parseTimeArgs(self, **kwargs):
        # parse arguments for a scheduled task
        ret = {}
        for i in TIME_TYPE_DEFS.keys():
            ret.update({i:[]})
        for key, value in kwargs.items():
            if ret.has_key(key):
                if type(value) is IntType:
                    ret[key] = (value, )
                elif type(value) is StringType:
                    ret[key] = (int(value),)
                elif type(value) in [ListType, TupleType]:
                    ret[key] = [int(v) for v in value]
            else:
                raise SchedulingError, "Invalid time keyword: %s" % key
        return ret

    security.declarePrivate( '_checkTimeArgs' )
    def _checkTimeArgs(self, args_dict):
        # check and time args and reformat (if needed)
        for key, value in args_dict.items():
            if not (value is None or self._isInValidRange(key, value)):
                raise SchedulingError, "Value '%s' for %s is invalid" % (str(value), key)
        minute = args_dict.get('minute')
        hour = args_dict.get('hour')
        dow = args_dict.get('dow')
        dom = args_dict.get('dom')
        month = args_dict.get('month')

        if type(minute) is IntType:
            minute = (minute,)
        elif type(minute) is not NoneType:
            minute = tuple(minute)
        else:
            minute = ()

        if minute is TupleType:
            ms = []
            for m in minute:
                ms.append(int(('%0.2d'%m)[-2:][0] + '0'))
            minute = tuple(ms)

        if type(hour) is IntType:
            hour = (hour,)
        elif type(hour) is not NoneType:
            hour = tuple(hour)
        else:
            hour = ()

        if type(dow) is IntType:
            dow = (dow,)
        elif type(dow) is not NoneType:
            dow = tuple(dow)
        else:
            dow = ()

        if type(dom) is IntType:
            dom = (dom,)
        elif type(dom) is not NoneType:
            dom = tuple(dom)
        else:
            dom = ()

        if type(month) is IntType:
            month = (month,)
        elif type(month) is not NoneType:
            month = tuple(month)
        else:
            month = ()
            
        return [minute, hour, dom, dow, month]

    def unregisterTask(self, path):
        """ """
        try:
            del self._tasks[path]
        except:
            pass

    def isRegisteredTask(self, path):
        return self._tasks.has_key(path)

    security.declarePrivate('_isInsideTimeSlot')
    def _isInsideTimeSlot(self, date, time_tuple):
        """ check whether a date is inside a given time tuple """

        dom = date.day()
        dow = date.dow()
        month = date.month()
        hour = date.hour()
        minute = date.minute()

        _minute, _hour, _dom, _dow, _month = time_tuple
        if (not _month) or (month in _month):
            if (not _dom) or (dom in _dom):
                if (not _dow) or (dow in _dow):
                    if (not _hour) or (hour in _hour):
                        if (not _minute) or (minute in _minute):
                            return 1
        return 0

    security.declarePrivate('findObjects')
    def findObjects(self, date):
        """ """
        traverse = self.unrestrictedTraverse
            
        result = []
        for path, time_data in self._tasks.items():
            if self._isInsideTimeSlot(date, time_data):
                try:
                    ob = traverse(path)
                except:
                    LOG("ZopeScheduler.findObjects", INFO, "Could not traverse path %s" % path)
                    ob = None
                if ob is not None:
                    result.append(ob)

        return result
        
    security.declarePrivate('process_timer')
    def process_timer(self, interval, tick, prev_tick, next_tick):
        """ process one message """
        year = tick.year()
        month = tick.month()
        day = tick.day()
        hour = tick.hour()
        minute = tick.minute()
                
        # decide process task or not
        date = DateTime('%s/%s/%s %s:%s:00'%(
            year, month, day, hour, ('%0.2ds'%minute)[0]+'0'))
        prev_delta = date - prev_tick
        if prev_delta > 0:
            if (tick - date) > prev_delta:
                return
        else:
            date = DateTime('%s/%s/%s %s:%s:00'%(
                year, month, day, hour, ('%0.2ds'%(minute+INTERVAL))[0]+'0'))
            next_delta = next_tick - date
            if next_delta > 0:
                if (date - tick) > next_delta:
                    return
            else:
                return

        # run tasks
        LOG('ZopeScheduler', INFO,
            'Process timer tick at %s\n'%tick)

        for ob in self.findObjects(date):
            LOG("Process task", INFO, "calling %s at %s" % (str(ob), str(date)))
            try:
                if ISchedulingTask.isImplementedBy(ob):
                    ob.process_task()
                elif callable(ob):
                    ob()
            except Exception, e:
		LOG('ZopeScheduler', ERROR, 'Process task error', error = sys.exc_info())

    security.declareProtected( Permissions.view, 'isActive' )
    def isActive( self ):
        """ """
        service = getTimerService(self)
        return service and not not service.listSubscriptions( self._event_stub )

    security.declareProtected( Permissions.manage_properties, 'activate' )
    def activate( self ):
        """ """
        service = getTimerService(self)
        if not service:
            raise ValueError, "Can't find event service!"

        service.subscribe( self )

    security.declareProtected( Permissions.manage_properties, 'deactivate' )
    def deactivate( self ):
        """ """
        service = getTimerService(self)
        if not service:
            raise ValueError, "Can't find event service!"

        service.unsubscribe( self )

    def manage_beforeDelete(self, item, container):
        self.deactivate()

    def manage_afterAdd( self, item, container ):
        self.activate()

    def manage_addTask(self, path, time_args=None, REQUEST=None):
        """ add a task (ZMI method) """
        msg = ''
        if not path:
            msg = 'You must specify path to subscriber'
            return MessageDialog(title='Scheduler Error',
                                 message=msg,
                                 action='manage_viewSubscriptions')
        else:
            try:
                ob = self.unrestrictedTraverse(path)
            except:
                ob = None

            if ob is None:
                msg = 'The path to subscriber seems to be invalid!'
                return MessageDialog(title='Scheduler Error',
                                     message=msg,
                                     action='manage_viewSubscriptions')

            else:
                if time_args is None:
                    msg = 'You must specify at least one date attribute!'
                    return MessageDialog(title='Scheduler Error',
                                         message=msg,
                                         action='manage_viewSubscriptions')
                else:
                    d = {}
                    for k, v in time_args.items():
                        d[k] = v
                    self.registerTask('/'.join(ob.getPhysicalPath()), **d)

        if REQUEST is not None:
            msg = 'Added subscriber with path %s' % (path,)
            return self.manage_viewSubscriptions(self, REQUEST, manage_tabs_message=msg)

    security.declareProtected( ScheduleTasks, 'manage_removeTasks' )
    def manage_removeTasks(self, paths=[], REQUEST=None):
        """ remove selected tasks (management method) """
        failed = []
        for path in paths:
            try:
                self.unregisterTask(path)
            except:
                failed.append(path)

        if REQUEST is not None:
            msg = 'Removed %s tasks!' % len(paths)
            if failed:
                msg = 'Failed to remove tasks at %s!' % (', '.join(failed))
            return self.manage_viewSubscriptions(self, REQUEST, manage_tabs_message=msg)


    security.declareProtected( ScheduleTasks, 'listTasks' )
    def listTasks(self):
        """ list available tasks """
        return self._tasks.items()

    security.declareProtected( ScheduleTasks, 'formatTasks' )
    def formatTask(self, task_date):
        """ format task properties for ZMI """
        _minute, _hour, _dom, _dow, _month = task_date
        dummy = 'Minutes: %s, Hours: %s, Day Of Month: %s, Day Of Week: %s, Month: %s'
        _minute = not _minute and '*' or '(%s)'%(', '.join([str(a) for a in _minute]))
        _hour = not _hour and '*' or '(%s)'%(', '.join([str(a) for a in _hour]))
        _dom = not _dom and '*' or '(%s)'%(', '.join([str(a) for a in _dom]))
        _dow = not _dow and '*' or '(%s)'%(', '.join([str(a) for a in _dow]))
        _month = not _month and '*' or '(%s)'%(', '.join([str(a) for a in _month]))
        ret = dummy % (_minute, _hour, _dom, _dow, _month)
        return ret

    security.declareProtected( ScheduleTasks, 'getTimeTypeDefs' )
    def getTimeTypeDefs(self):
        """ return TIME_TYPE_DEFS suitable for manage_viewSubscriptions """
        # keep keys in order
        keys = ['minute', 'hour', 'dom', 'dow', 'month']
        ret = []
        for key in keys:
            d = {'type_name': key}
            ttype = TIME_TYPE_DEFS[key]
            steps = range(ttype['min'], ttype['max']+1)
            if ttype.has_key('fraction'):
                steps = [s for s in steps if s % ttype['fraction'] == 0]
            d['steps'] = steps
            d['title'] = ttype['name']
            d['type_unit'] = ttype['unit']
            ret.append(d)
        return ret

    security.declarePrivate( '_isInValidRange' )
    def _isInValidRange(self, time_type, value):
        # check if a given value is in valid time range
        time_def = TIME_TYPE_DEFS[time_type]
        _min = time_def['min']
        _max = time_def['max']
        if type(value) not in [ListType, TupleType]:
            value = [value]
        for v in value:
            if v > _max or v < _min:
                return 0
        return 1

InitializeClass(ZopeScheduler)
